package SecureDistributedKeyGenerationForDiscreteLogBasedCryptosystem;

import Communication.User;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import JointFeldmanProtocol.DistributedKeyGeneration;
import ShamirSecretSharing.Polynomial;
import com.google.protobuf.ByteString;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 3/16/2016.
 */
public class SecureDistributedKeyGeneration extends DistributedKeyGeneration {

    private VerifiableSecretSharing verifiableSecretSharing;
    private final BigInteger h;
    private SecureDistributedKeyGenerationParty[] parties;

    public SecureDistributedKeyGeneration(int t, int n, BigInteger zi, Random random, BigInteger q, BigInteger g
            , BigInteger h, Group<BigInteger> group, int id) {
        super(t, n, zi, random, q, g, group, id);
        this.h = h;
        BigInteger r = new BigInteger(q.bitLength(),random).mod(q);
        this.verifiableSecretSharing = new VerifiableSecretSharing(t,n,r,random,q,h,group);
        this.parties = new SecureDistributedKeyGenerationParty[n];
        for (int i = 1; i <= n ; i++){
            this.parties[i - 1] = new SecureDistributedKeyGenerationParty(i,n,t);
        }
        this.parties[id - 1].share = getShare(id);
        this.parties[id - 1].shareT = verifiableSecretSharing.getShare(id);
        super.setParties(parties);
    }

    protected SecureDistributedKeyGenerationParty[] getParties(){
        return parties;
    }

    protected void setParties(SecureDistributedKeyGenerationParty[] parties) {
        super.setParties(parties);
        this.parties = parties;
    }


    @Override
    public void sendSecret(User user, int j) {
        Polynomial.Point secret = getShare(j);
        Polynomial.Point secretT = verifiableSecretSharing.getShare(j);
        DKGMessages.DoubleSecretMessage doubleSecretMessage = doubleSecretMessage(id,j,secret,secretT);
        user.send(j, DKGMessages.Mail.Type.SECRET, doubleSecretMessage);
    }


    @Override
    public boolean isValidSecret(int i){
        SecureDistributedKeyGenerationParty party = parties[i - 1];
        return isValidSecret(party.share, party.shareT, party.verifiableValues, id);
    }

    /**
     * @param secret
     * @param secretT
     * @param verificationValues
     * @param j
     * @return verify(j,verificationValues,group) == (g ^ secret.y) * (h ^ secretT.y) mod q
     */
    public boolean isValidSecret(Polynomial.Point secret,Polynomial.Point secretT, BigInteger[] verificationValues, int j){
        try {
            BigInteger v = verify(j, verificationValues, group);
            BigInteger exp = group.add(group.multiply(g, secret.y), group.multiply(h, secretT.y));
            return exp.equals(v);
        }
        catch (NullPointerException e){
            return false;
        }
    }

    private void broadcastComplaint(User user,Polynomial.Point secret,Polynomial.Point secretT,int i){
        DKGMessages.DoubleSecretMessage complaint = doubleSecretMessage(i,id,secret,secretT);
        user.broadcast(DKGMessages.Mail.Type.COMPLAINT,complaint);
    }

    /**
     * stage4.3 according to the protocol
     *  if check fails for index i, Pj
     */
    public void broadcastComplaints(User user, Set<Integer> QUAL){
        SecureDistributedKeyGenerationParty party;
        for (int i : QUAL) {
            party = parties[i - 1];
            if (i != id) {
                if (!super.isValidSecret(party.share, party.commitments, id)) {
                    broadcastComplaint(user, party.share, party.shareT, i);
                }
            }
        }
    }

    public void broadcastVerificationValues(User user){
        BigInteger[] verificationValues = new BigInteger[t + 1];
        BigInteger[] hBaseCommitments = verifiableSecretSharing.getCommitmentsArray();
        for (int k = 0 ; k < verificationValues.length ; k++){
            verificationValues[k] = group.add(commitmentsArray[k],hBaseCommitments[k]);
        }
        broadcastCommitments(user,verificationValues);
    }

    private DKGMessages.DoubleSecretMessage doubleSecretMessage(int i,int j,Polynomial.Point secret, Polynomial.Point secretT){
        DKGMessages.DoubleSecretMessage doubleSecretMessage = DKGMessages.DoubleSecretMessage.newBuilder()
                .setI(i)
                .setJ(j)
                .setSecret(ByteString.copyFrom(secret.y.toByteArray()))
                .setSecretT(ByteString.copyFrom(secretT.y.toByteArray()))
                .build();
        return doubleSecretMessage;
    }

    @Override
    public void broadcastComplaintAnswer(User user, int j) {
        DKGMessages.DoubleSecretMessage answer = doubleSecretMessage(id,j,getShare(j)
                ,verifiableSecretSharing.getShare(j));
        user.broadcast(DKGMessages.Mail.Type.ANSWER,answer);
    }

    public void broadcastAnswer(User user,Polynomial.Point secret,Polynomial.Point secretT,int i){
        DKGMessages.DoubleSecretMessage complaint = doubleSecretMessage(i,id,secret,secretT);
        user.broadcast(DKGMessages.Mail.Type.ANSWER,complaint);
    }

    public BigInteger getH() {
        return h;
    }
}
